/*
 * Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  Further, any license provided herein,
 * whether implied or otherwise, is limited to this program in accordance
 * with the express provisions of the GNU General Public License.  Patent
 * licenses, if any, provided herein do not apply to combinations of this
 * program with other product or programs, or any other product whatsoever.
 * This program is distributed without any warranty that the program is
 * delivered free of the rightful claim of any third person by way of
 * infringement or the like.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */

/*
 * Author:      Dean Johnson (dtj@sgi.com)
 *
 * Purpose:     This is the plugin handler
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/dir.h>

#ifdef CRAY
typedef long	Boolean;
#else
typedef char	Boolean;
#endif

#ifndef	True
#define	True	1
#define	False	0
#endif	/*True*/

typedef	struct {
	char	*name;
	char	*filename;
	void	*handle;
	char	*(*Name)(void);
	void	(*PreXHook)(void);
	void	(*PostXHook)(void);
	void	(*DeleteHook)(void);
	} Plugin;

Plugin	**PluginList = NULL;
int	PluginListCnt = 0;
Boolean	calledPreHook = False;
Boolean	calledPostHook = False;

void
PIM_DumpTables(void)
{
	for (int i = 0; i < PluginListCnt; i++) {
		fprintf(stderr,"------ Plug in table ---------\n");
		fprintf(stderr,"filename:\t0x%08x %s\n",(unsigned int) PluginList[i]->filename,
			(PluginList[i]->filename) ? PluginList[i]->filename : "NULL");
		fprintf(stderr,"\tname:\t0x%08x %s\n",(unsigned int) PluginList[i]->name,
			(PluginList[i]->name) ? PluginList[i]->name : "NULL");
		fprintf(stderr,"\thandle:\t0x%08x\n",(unsigned int) PluginList[i]->handle);
		fprintf(stderr,"\tName():\t0x%08x\n",(unsigned int) PluginList[i]->Name);
		fprintf(stderr,"\tPreXHook():\t0x%08x\n",(unsigned int) PluginList[i]->PreXHook);
		fprintf(stderr,"\tPostXHook():\t0x%08x\n",(unsigned int) PluginList[i]->PostXHook);
	}
}

static Boolean
NukePlugin(char *name)
{
	int	i;
	int	dead = 0;

	if (name == NULL) return(False);
	for (i = 0; i < PluginListCnt; i++) {
		if (PluginList[i]->name == NULL) continue; // already deleted
		if (!strcmp(PluginList[i]->name,name)) {
			if (PluginList[i]->DeleteHook != NULL) {
				/*
				 *	Call the destructor!
				 */
				(*PluginList[i]->DeleteHook)();
			}
			free(PluginList[i]->name);
			PluginList[i]->name = NULL;
			dead++;
		}
	}
	return(dead ? True : False);
}

static Plugin *
NewPlugin(char *filename,void *handle) 
{
	Plugin	*pi;

	pi = (Plugin *) malloc(sizeof(Plugin));
	pi->handle = handle;
	pi->filename = (filename == NULL) ? NULL : strdup(filename);
	pi->name = NULL;

	/*
	 *
	 */
	pi->Name = (char *(*)(void)) dlsym(handle,"PIM_Name");
	if (pi->Name == NULL) {
		//	no PIM_Name(), so this isn't a legit plugin
		free(pi);
		return(NULL);
	}
	pi->name = strdup((*pi->Name)());
	pi->PreXHook   = (void (*)(void)) dlsym(handle,"PIM_PreXHook");
	pi->PostXHook  = (void (*)(void)) dlsym(handle,"PIM_PostXHook");
	pi->DeleteHook = (void (*)(void)) dlsym(handle,"PIM_DeleteHook");

	/*
	 *	Stash it away now
	 */
	if (!PluginListCnt) {
		PluginListCnt++;
		PluginList = (Plugin **) malloc(sizeof(Plugin *));
	}
	else	{
		PluginListCnt++;
		PluginList = (Plugin **) realloc(PluginList,sizeof(Plugin *) 
							    * PluginListCnt);
	}
	PluginList[PluginListCnt-1] = pi;
	return(pi);
}

void
PIM_RunPreXHook(void)
{
	for (int i = 0; i < PluginListCnt; i++) {
		if (PluginList[i]->name == NULL) continue;	// deleted
		if (PluginList[i]->PreXHook != NULL) {
#ifdef	notdef
			fprintf(stderr,"%s: calling pre hook\n",
				PluginList[i]->filename);
#endif	/*notdef*/
			(*PluginList[i]->PreXHook)();
		}
	}
	calledPreHook = True;
}

void
PIM_RunPostXHook(void)
{
	for (int i = 0; i < PluginListCnt; i++) {
		if (PluginList[i]->name == NULL) continue;	// deleted
		if (PluginList[i]->PostXHook != NULL) {
#ifdef	notdef
			fprintf(stderr,"%s: calling post hook\n",
				PluginList[i]->filename);
#endif	/*notdef*/
			(*PluginList[i]->PostXHook)();
		}
	}
	calledPostHook = True;
}

static int
CheckOutPluginDirectory(char *dirname)
{
	DIR	*theDir;
	char	thePath[1024];
	int	thePathLen;
	int	plugCount = 0;
	struct direct	*theEntry;
	void	*theHandle;
	Plugin	*pi;

	if ((theDir = opendir(dirname)) == NULL) {
		// can't open plugin's directory
//		fprintf(stderr,"Unable to open plugin's directory \"%s\"!\n",
//			dirname);
		return(0);
	}
	strcpy(thePath,dirname);
	strcat(thePath,"/");
	thePathLen = strlen(thePath);
	while ((theEntry = readdir(theDir)) != NULL) {
		if (   !strcmp(theEntry->d_name,".")
		    || !strcmp(theEntry->d_name,".."))
			continue;
		strcpy(&thePath[thePathLen],theEntry->d_name);
//		fprintf(stderr,"looking at plugin %s\n",thePath);
		if ((theHandle = dlopen(thePath,RTLD_LAZY)) == NULL) {
//			fprintf(stderr,"%s\n",dlerror());
			continue;
		}
		pi = NewPlugin(thePath,theHandle);
		if (pi == NULL) continue;
		if (!plugCount) {
			fprintf(stderr,"Plugins: ");
		}
		plugCount++;
		fprintf(stderr,"%s  ",theEntry->d_name);
	}
	if (plugCount)
		fprintf(stderr,"\n");
	closedir(theDir);
	return(plugCount);
}

void
PIM_Initialize(void)
{
	char	*dirname = NULL;
	int	plugCount = 0;

	/*
	 *	Check out the system plugins
	 */

	plugCount += CheckOutPluginDirectory("/usr/lib/vacm/hoover/plugins");

	/*
	 *	Check out the users plugins in his home directory
	 */
	if ((dirname = getenv("HOME")) != NULL) {
		char	buf[1024];
		strcpy(buf,dirname);
		strcat(buf,"/.hoover/plugins");
		plugCount += CheckOutPluginDirectory(buf);
	}

	/*
	 *	Check out the users specified plugins
	 */

	if ((dirname = getenv("HOOVER_PLUGINS")) == NULL) {
		// no plugin's indicated
#ifdef	notdef
		fprintf(stderr,"No plugin's directory!\n");
#endif	/*notdef*/
		return;
	}
	plugCount += CheckOutPluginDirectory(dirname);
}

int
PIM_AddPlugin(char *theFilename)
{
	Plugin	*pi;

	void	*theHandle = dlopen(theFilename,RTLD_LAZY);
	if (theHandle == NULL) return(0);
	if ((pi = NewPlugin(theFilename,theHandle)) == NULL) return(0);

	if (calledPreHook && (pi->PreXHook != NULL)) {
		(*pi->PreXHook)();
	}
	if (calledPostHook && (pi->PostXHook != NULL)) {
		(*pi->PostXHook)();
	}
	return(1);
}

void     
PIM_DeletePlugin(char *name)
{
	(void) NukePlugin(name);
}
