/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 *
 * Copyright (C) 2001, San Mehat, Zac Sprackett and VA Linux Systems
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.  Redistributions
 * in binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * Neither the name of the VA Linux Systems, Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <netdb.h>
#ifndef __sgi
#include <linux/types.h>
#include <linux/serial.h>
#endif
#include <libloose.h>
#include <skel_module_harness.h>
#include <skel_slave_context.h>
#include <prototypes.h>
#include <skel_slave_context.h>
#include <pthread.h>

static void	thread_cleanup(void *);
static void	process_msg_from_slave(CONTEXT *);
static int	recv_msg(int , char **, __uint32_t *);
static void	process_skel_ipc(CONTEXT *, __uint32_t, int, char *, char *);
static int	send_msg(int , char *);

/*
 * This stuff is so that a thread can 'peek' at another threads context
 * structure (to check for duplicate serial ports, etc)
 */
extern pthread_mutex_t	context_mutex;
extern CONTEXT					*context_head;

static void thread_cleanup(void *data)
{
	CONTEXT *ctx = (CONTEXT *) data;
	
#ifdef THREAD_DEBUG
	printf("Thread %s cleaning up\n",ctx->node_id);
#endif
#ifdef THREAD_DEBUG
	printf("Thread %s disengaging\n",ctx->node_id);
#endif
	if (ctx->words)
		fclose(ctx->words);
#ifdef THREAD_DEBUG
	printf("Thread %s disengaged\n",ctx->node_id);
	printf("Thread %s clean\n",ctx->node_id);
#endif
	ctx->refcount--;
}

void *thread_start(void *data)
{
	struct _pthread_cleanup_buffer 	_buff;
	CONTEXT													*ctx = (CONTEXT *) data;
	struct timeval									to;
	fd_set													read_fds;
	int															max, rc;
	char														log_msg[255];

#ifdef THREAD_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s starting (pid %d)",
					ctx->node_id,
					getpid());
	send_msg(ctx->thread_ipc_write, log_msg);
#endif
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	_pthread_cleanup_push(&_buff, thread_cleanup, (void *) ctx);
	ctx->refcount++;

	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(ctx->thread_ipc_read, &read_fds);
		max = ctx->thread_ipc_read;

		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s select() failed (%s)",
							ctx->node_id,
							strerror(errno));
			send_msg(ctx->thread_ipc_write, log_msg);
			pthread_testcancel();
			continue;
			}
		if (!rc)
			{
			pthread_testcancel();
			continue;
			}
		if (FD_ISSET(ctx->thread_ipc_read, &read_fds))
			{
			process_msg_from_slave(ctx);
			}
		pthread_testcancel();
		}
}

static void process_msg_from_slave(CONTEXT *ctx)
{
	char				*msg_buffer, *cmd, *node_id, *bword, *saved_buffer;
	char				*ipc_cmd, *ipc_args, *variable, *value;
	int					len, rc, client_fd;
	__uint32_t	job_id;
	char				*last = (char *) 0x00;
	char				log_msg[255];

	if ((rc = recv_msg(ctx->thread_ipc_read, &msg_buffer, &len)) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to read msg from slave (%s)",
					 	ctx->node_id,
					 	strerror(-rc));
		send_msg(ctx->thread_ipc_write, log_msg);
		pthread_exit(NULL);
		return;
		}
	saved_buffer = malloc(len);
	memcpy(saved_buffer, msg_buffer, len);

#if 0
	printf("Thread %s received msg (%s) from slave\n",
				 ctx->node_id,
				 msg_buffer);
#endif
	if (!(cmd = strtok_r(msg_buffer, ":", &last)))
		goto out;
	if (!(node_id = strtok_r(NULL, ":", &last)))
		goto out;
	if (!strcasecmp(cmd,"NODE_IPC"))
		{
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		job_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(bword);
		if (!(ipc_cmd = strtok_r(NULL, ":", &last)))
			goto out;
		ipc_args = &saved_buffer[((ipc_cmd - msg_buffer)+strlen(ipc_cmd)+1)];
		process_skel_ipc(ctx, job_id, client_fd, ipc_cmd, ipc_args);
		}
	else if (!strcasecmp(cmd, "NODE_VAR"))
		{
		if (!(variable = strtok_r(NULL, ":", &last)))
			goto out;

		/* BEGIN San mod 07/12/00 */
		value = &saved_buffer[((variable - msg_buffer)+(strlen(variable)+1))];
		/* END San mod 07/12/00 */
#error variables not adjusted for module
		if (!strcasecmp(variable,"DEVICE_ADDRESS"))
			{
			goto out;
			}
		else if (!strcasecmp(variable,"BAUDRATE"))
			{
			goto out;
			}
		}
	else if (!strcasecmp(cmd, "CLIENT_DIS"))
		{
		char								*fd;
		int									client_fd;
		char								clmsg[255];

		/* Called when a client disconnects from nexxus */
		if (!(fd = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(fd);
#error client disconnection code not adjusted for module
		goto out;
		}
	else if (!strcasecmp(cmd, "NODE_RENAME"))
		{
		char	*new_name;
	
		/* Called when a node is renamed */
		if (!(new_name = strtok_r(NULL, ":", &last)))
			goto out;
#error node rename code not adjusted for module
		goto out;
		}
	else
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s received unsupported cmd from slave (%s)",
					 	ctx->node_id,
					 	cmd);
		send_msg(ctx->thread_ipc_write, log_msg);
		}

out:
	free(saved_buffer);
	free(msg_buffer);
}

/*
 * Process an IPC message
 */
static void process_skel_ipc(CONTEXT			*ctx,
															__uint32_t	job_id,
															int					client_fd,
															char				*ipc_cmd,
															char				*ipc_args)
{
	char					resp_msg[255];
	char					*last = (char *) 0x00;
	char					log_msg[255];

#error IPC message processing not adjusted for module
	if (!strcasecmp(ipc_cmd, "IPC_MSG1"))
		{
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "IPC_MSG2"))
		{
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else
		{
		sprintf(resp_msg,
						"%d:%d:FOR:JOB_ERROR:%s",
						job_id,
						client_fd,
						strerror(EBADRQC));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}

out_ok:
	return;

out_malformatted:
	sprintf(resp_msg,
					"%d:%d:FOR:JOB_ERROR:%s",
					job_id,
					client_fd,
					strerror(EINVAL));
	send_msg(ctx->thread_ipc_write, resp_msg);
	return;
}
	
/*
 * Receive a message. This function is threadsafe
 */
static int recv_msg(int fd, char **buffer, __uint32_t *len)
{
  __uint32_t  net_len;
	__uint32_t	brtr;
	char				*offset;
  int         rc;

  if ((rc = read(fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
	else if (!rc)
		return(-1);
  *len = ntohl(net_len);

  if (!(*buffer = malloc(*len)))
    return(-ENOMEM);
  memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr)
		{
  	if ((rc = read(fd, (void *) offset, brtr)) < 0)
    	return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset += rc;
		}
  return(0);
}

/*
 * Send a message. This function is threadsafe
 */
static int send_msg(int fd, char *msg)
{
  __uint32_t  len;
  __uint32_t  net_len;
  int         rc;

	if (fd == -1)
		return(-EBADF);
  len = strlen(msg)+1;
  net_len = htonl(len);
  if ((rc = write (fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
  if ((rc = write (fd, (void *) msg, len)) < 0)
    return(rc);
  return(0);
}
